#!/usr/bin/env python

import os
import subprocess
import shutil
import sys
import shutil
import re
try:
    from gi.repository import Notify
    DISABLE_LIBNOTIFY=False
except ImportError:
    DISABLE_LIBNOTIFY=True

learners = ['lfit',
            'pasula']

# update_progress() : Displays or updates a console progress bar
## Accepts a float between 0 and 1. Any int will be converted to a float.
## A value under 0 represents a 'halt'.
## A value at 1 or bigger represents 100%
def update_progress(progress):
    barLength = 10 # Modify this to change the length of the progress bar
    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(barLength*progress))
    text = "\rPercent: [{0}] {1}% {2}".format( "#"*block + "-"*(barLength-block), progress*100, status)
    sys.stdout.write(text)
    sys.stdout.flush()


def get_bin_path(current_dir):
    bin_path = ''
    if os.path.exists(current_dir + '/bin'):
        bin_path = current_dir + '/bin/'
    elif os.path.exists(current_dir + '/../bin'):
        bin_path = current_dir + '/../bin/'
    elif os.path.exists(current_dir + '/../../bin'):
        bin_path = current_dir + '/../../bin/'
    else:
        bin_path = current_dir + '/../../../bin'
    return os.path.realpath(bin_path)

def execute_command(command):
    res = subprocess.call(command, shell=True)
    if (res != 0):
        print "Unexpected error"
        if not DISABLE_LIBNOTIFY:
            Notify.init("REX-D")
            notify_msg = Notify.Notification.new("REX-D","Unexpected error executing " + command + "! ","dialog-error")
            notify_msg.show ()
        raise RuntimeError

def read_error_probability():
    differences_file="tempo_differences.txt"
    f = open(differences_file, 'r')
    for line in f:
        if "Differences are" in line:
            return float(line.split("are")[1])

def do_single_benchmark(learner, num_transitions, number_simulations):
    current_dir=os.getcwd()
    binary_path = get_bin_path(current_dir) + '/utils_rexd'
    command_generate_transitions = binary_path + " transitions " + str(num_transitions) + " > ./tempo_transitions.txt"
    command_learn = binary_path + " learn " + learner + " > ./tempo_learn.txt"
    command_test_probability = binary_path + " evaluate_rules learned_rules.ppddl " + str(number_simulations) + " > ./tempo_differences.txt"
    #print "Generating transitions"
    execute_command(command_generate_transitions)
    #print "Learning"
    finished = False
    while not finished:
        try:
            execute_command(command_learn)
            finished = True
        except RuntimeError:
            print "Retrying learning..."

    #print "Checking rules"
    #shutil.copy2(current_dir + '/' + 'learned_rules.ppddl', 'domain.pddl')
    execute_command(command_test_probability)

    error_prob = read_error_probability()

    return error_prob

def do_set_benchmarks(learner, min_transitions, max_transitions, num_repetitions, number_simulations):
    out_file = open('learning_results.txt','w')
    result_vector = learner + '=['
    for number_transitions in range(min_transitions,max_transitions+1):
        name_current_iteration = learner + '_exp' + str(number_transitions)
        print "Number transitions: " + str(number_transitions)
        average_prob = 0.0
        out_file.write(name_current_iteration + '_m' + '=[')
        for i in range(0,num_repetitions):
            update_progress(float(i)/float(num_repetitions))
            new_prob = do_single_benchmark(learner, number_transitions, number_simulations)
            average_prob = average_prob + new_prob
            out_file.write(str(new_prob))
            if i < (num_repetitions-1):
                out_file.write(",")
        out_file.write('];\n')
        out_file.write(name_current_iteration + ' = mean(' + name_current_iteration + '_m);\n')
        out_file.write(name_current_iteration + '_v' + ' = var(' + name_current_iteration + '_m);\n')
        out_file.write(name_current_iteration + '_E' + ' = std(' + name_current_iteration + '_m);\n')
        out_file.flush() # save just in case
        update_progress(1)

        # Show in terminal the average for current iteration
        average_prob = average_prob / num_repetitions
        print "Got avg prob: " + str(average_prob)

        # Write to result_vector string
        result_vector += name_current_iteration
        if (number_transitions < max_transitions):
            result_vector += ','

    out_file.write(result_vector + '];\n')
    out_file.close()


def main():
    if len(sys.argv) < 6:
        print "Not enough arguments."
        print "learner_bench.py learner min_transitions max_transitions num_repetitions num_simulations"
    else:
        do_set_benchmarks(sys.argv[1],int(sys.argv[2]),int(sys.argv[3]), int(sys.argv[4]), int(sys.argv[5]))

if __name__ == "__main__":
    main()
