/*
 * LFIT probabilistic postprocessing
 * Copyright (C) 2015  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef REXD_PROBABILISTIC_LFIT_H
#define REXD_PROBABILISTIC_LFIT_H 1

#include <boost/optional.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/lambda/lambda.hpp>
#include <symbolic/reader_base.h>
#include <symbolic/domain.h>
#include <rexd/config_reader.h>
#include <lfit/lfit_types.h>
#include <lfit/subsumption_graph.h>
#include <lfit/lfit_conflict_solver.h>

/**
 * \class ProbabilisticLFITAnalyzer
 * \brief Chooses the best subset of LFIT rules to represent a domain
 * 
 */
class ProbabilisticLFITAnalyzer
{
public:
    ProbabilisticLFITAnalyzer(const uint debug, const TransitionGroup& transitions, const ConfigReader& config_reader):
        debug(debug),
        transitions(transitions),
        config_reader(config_reader)
    {};
private:
    uint debug; // 0 -> nothing, 1 -> basic algorithm, 2 -> data read, 3 -> data parsed, 4 -> details
    TransitionGroup transitions;
    ConfigReader config_reader;
    
    float get_score(const lfit::RuleWithCoverageSet& ruleset);
    RuleSet select_rules_for_each_head(const lfit::MapHeadRules& map_head_rules, const TransitionGroup& action_transitions);
    lfit::RuleWithCoverageSet select_rules_for_head(const lfit::RuleWithCoverageSet& action_head_rules,
                                                 const TransitionGroup& transitions);
    
    lfit::RuleWithCoverageSet select_best_ruleset(const lfit::RuleWithCoverageSet& current_candidates,
                                                  const TransitionGroup& transitions,
                                                  const int verbose
                                                 );
    
    bool remove_conflicting_leafs(const TransitionGroup& head_transitions, const lfit::RuleWithCoverageSet& rules_to_check, 
                                  lfit::RuleSubsumptionGraph& subsumption_graph, 
                                  lfit::RuleWithCoverageSet& best_operators, double& best_candidate_score);
    
    boost::optional<RuleSet> check_inverted_rules(const lfit::MapHeadRules& learned_rule_sets, const Predicate<TypedAtom>& current_head, const TransitionGroup& head_transitions);
    
    TransitionGroup obtain_head_transitions(const TransitionGroup& transitions, const Predicate<TypedAtom>& head);
    
public:
    RuleSetPtr select_rules(const std::set<lfit::RuleWithCoverage>& lfit_rules);
    
};


#endif