#include "learner.h"


Learner::Learner() {
}


RuleSetPtr Learner::learn_action_rules(const RuleSet& previous_rules, const TransitionGroup& transitions, const TypedAtom action)
{
    RuleSetPtr removed_rules, added_rules;
    return learn_action_rules(previous_rules, transitions, action, removed_rules, added_rules);
}

RuleSetPtr Learner::learn_action_rules(const RuleSet& previous_rules, const TransitionGroup& transitions, const TypedAtom& action, RuleSetPtr& removed_rules_ptr, RuleSetPtr& added_rules_ptr)
{
    int DEBUG = 0;
    TypedAtom symbolic_action(action);
    symbolic_action.make_symbolic();
    TransitionGroup transitions_action = transitions.get_action_transitions(symbolic_action);
    if (DEBUG > 0) {
        LOG(INFO) << "Learner::learn_action_rules - " << "Starting learning...";
    }
    if (DEBUG > 1) {
        LOG(INFO) << "Action " << action << " transitions are " << "\n" << transitions_action;
        added_rules_ptr = learn(transitions_action);
    }
    else {
        added_rules_ptr = learn_quiet(transitions_action);
    }
    if (DEBUG > 0) {
        LOG(INFO) << "Learner::learn_action_rules - " << "Finished learning.";
    }
    
    // for each rule, set correct param types
    BOOST_FOREACH(RulePtr& rule_ptr, *added_rules_ptr) {
        for (size_t i = 0; i < action.params.size(); ++i) {
            rule_ptr->change_param_types(rule_ptr->get_param_names(), rule_ptr->get_param_types());
        }
    }

    RuleSetPtr result_rules = boost::make_shared<RuleSet>(RuleSet(previous_rules));

    RuleSet removed_rules = result_rules->remove_action_rules(symbolic_action);
    removed_rules_ptr = boost::make_shared<RuleSet>(removed_rules);
    
    // maintain same identifiers as much as possible in newer rules
    added_rules_ptr->set_identifiers(removed_rules_ptr);

    // Add new rules to current rules
    result_rules->add_rules(*added_rules_ptr);
    
    return result_rules;
}


