/*
 * Scenario class that has all information needed to create planning models to be used by REX-D.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "rexd_scenario.h"

RexdScenario::RexdScenario(const RexdScenario& rexd_scenario): 
    Scenario(rexd_scenario),
    use_rmax(rexd_scenario.use_rmax),
    use_vmin(rexd_scenario.use_vmin),
    vmin_value(rexd_scenario.vmin_value)
{
    if (use_rmax) {
        set_rmax(rexd_scenario.transitions_ptr);
    }
}

RexdScenario::RexdScenario(const Scenario& scenario): 
    Scenario(scenario),
    use_rmax(false),
    use_vmin(false)
{
    reset_vmin_value();
}


void RexdScenario::set_rmax(const TransitionGroupPtr transitions_ptr_)
{
    use_rmax = false;
    if (config_reader.get_variable<bool>("use_rmax")) {
        this->use_rmax = true;
        this->rmax_reward = config_reader.get_variable<float>("rmax_reward");
        this->explore_threshold = config_reader.get_variable<uint>("exploration_known_threshold");
        this->transitions_ptr = transitions_ptr_;
    }
}


void RexdScenario::set_vmin()
{
    use_vmin = false;
    if (config_reader.get_variable<bool>("use_vmin")) {
        if (!config_reader.get_variable<bool>("final_reward_action"))
            LOG(WARNING) << "V-MIN requires final_reward_action to be enabled.";
        this->use_vmin = true;
    }
}

