/*
 * Class representing a state (as a set of predicates).
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef INTELLACT_STATE_H
#define INTELLACT_STATE_H

#include <iterator>
#include "symbolic/predicates.h"

/*!
\class State
\brief Class defining the current scenario state.

At this moment it is just a PredicateGroup. In the future it may be enriched with partial observability.
*/
class State : public PredicateGroup<TypedAtom>
{
private:
    PredicateGroup<TypedAtom> constant_predicates;
    
public:
    PPDDLObjectManager ppddl_object_manager;
    
    State() {};
    explicit State(const std::string& state_string): PredicateGroup<TypedAtom>(state_string) {};
    State(const PredicateGroup<TypedAtom>& pred_group, const PPDDLObjectManager& pddl_object_manager): 
        PredicateGroup<TypedAtom>(pred_group),
        ppddl_object_manager(pddl_object_manager)
    {};
    State(const State& state_): 
        PredicateGroup<TypedAtom>(state_),
        constant_predicates(state_.constant_predicates),
        ppddl_object_manager(state_.ppddl_object_manager)
        { };
        
    void clear();
    
    void clear_nonconstant_predicates();
    
    void set_constants(const PredicateGroup<TypedAtom>& new_constant_predicates);
    const PredicateGroup<TypedAtom>& get_constants() const { return constant_predicates; };
    PredicateGroup<TypedAtom> get_non_constants() const;

//     void read_from_file(std::string file_path);

    void write_to_file(std::string file_path) const;
    
    void update_predicate(Predicate<TypedAtom> new_predicate);
    
    /*!
    \brief Removes special predicates such as +-finished() and +-started()
    */
    void remove_special_predicates();
    
//     void read_ppddl_objects_from_string(std::string ppddl_objects_str);
    
//     PpddlTypeObjectMap get_ppddl_objects_per_type() const { return ppddl_objects_per_type; };
    
//     PpddlObjectTypeMap get_ppddl_objects_types() const { return ppddl_objects_types; };
    
    // TODO C++11 const method -> use map::at() which throws exceptions
//     TypedAtom get_typed_atom(const Atom& symbol);
    
    // TODO C++11 const method -> use map::at() which throws exceptions
//     std::string get_object_type(const std::string& object);

private:
    
    
};

typedef boost::shared_ptr<State> StatePtr;
typedef boost::shared_ptr<const State> StateConstPtr;


#endif
